######################################################################
# Author: Michael Hoehle <http://www.stat.uni-muenchen.de/~hoehle>
# Date:   23 October 2005
#
# Description:
# This R file is used to visualize the results of an epidemic
# ran by simsellke.sh. See the README.txt for a sequence
# of commands.
######################################################################

######################################################################
# Convert the .all file to one giving the number of E,I,S in each
# cell. Take object and initialize ts param
#
# Params:
#  obj - Ladybug Object
######################################################################

init.spacetime <- function(obj) {
  #Initial number of susceptibles, exposed, infectious and recovered
  #The inital numbers (infectives are ignored, coz they will appear
  #as exposed first (if run with option e0)
  S <- array(NA,dim=c(obj$noX,obj$noY,dim(obj$epi)[1]+1))
  E <- array(NA,dim=c(obj$noX,obj$noY,dim(obj$epi)[1]+1))
  I <- array(NA,dim=c(obj$noX,obj$noY,dim(obj$epi)[1]+1))

  #Start values
  S[,,1] <- matrix(obj$setup$S+obj$setup$I,nrow=obj$noX,ncol=obj$noY)
  E[,,1] <- matrix(0,nrow=obj$noX,ncol=obj$noY)
  I[,,1] <- matrix(0,nrow=obj$noX,ncol=obj$noY)

  for (i in 1:dim(obj$epi)[1]) {
    type <- obj$epi[i,]$type
    #Location of the event
    evx <- obj$epi[i,]$x
    evy <- obj$epi[i,]$y

    #Loop over all cells
    for (x in 1:obj$noX) {
      for (y in 1:obj$noY) {
        #Carry last value forward
        S[x,y,i+1] <- S[x,y,i]
        E[x,y,i+1] <- E[x,y,i]
        I[x,y,i+1] <- I[x,y,i]
        
        #If the event is in the current cell
        if ((evx == x) & (evy == y)) {
          if (type == "E") {
            S[x,y,i+1] <- S[x,y,i+1] - 1 ; E[x,y,i+1] <- E[x,y,i+1] + 1
          } else if (type == "I") {
            E[x,y,i+1] <- E[x,y,i+1] - 1 ; I[x,y,i+1] <- I[x,y,i+1] + 1
          } else if (type == "R") {
            I[x,y,i+1] <- I[x,y,i+1] - 1
          }
        }
      }
    }
  }
  obj$ts <- list(S=S,E=E,I=I)
  return(obj)
}


######################################################################
# Convert epi data to a suitable format
#
# Params:
#  obj - ladybug object
init <- function(obj) {
  #Initial number of susceptibles, exposed, infectious and recovered
  #The inital numbers (infectives are ignored, coz they will appear
  #as exposed first (if run with option e0)
  S <- matrix(obj$setup$S+obj$setup$I,nrow=obj$noX,ncol=obj$noY)
  E <- matrix(0,nrow=obj$noX,ncol=obj$noY)
  I <- matrix(0,nrow=obj$noX,ncol=obj$noY)

  #Add columns for E,I,S
  obj$epi$S <- NA
  obj$epi$E <- NA
  obj$epi$I <- NA
  
  for (i in 1:dim(obj$epi)[1]) {
    type <- obj$epi[i,]$type
    x <- obj$epi[i,]$x
    y <- obj$epi[i,]$y

    #Action depends on type
    if (type == "E") {
      S[x,y] <- S[x,y] - 1 ; E[x,y] <- E[x,y] + 1
    } else if (type == "I") {
      E[x,y] <- E[x,y] - 1 ; I[x,y] <- I[x,y] + 1
    } else if (type == "R") {
      I[x,y] <- I[x,y] - 1
    }
    ##Set the events
    obj$epi[i,c("S","E","I")] <- c(S[x,y],E[x,y],I[x,y])
  }
  return(obj)
}

######################################################################
# A function to plot the course of the epidemic for each cell.
#
# Params:
#  obj - ladybug object 
######################################################################

plot.epilocal <- function(obj) {
  #Make the appropriate call to par
  par(mfrow=c(obj$noX,obj$noY),mar=c(3,1,3,1))
  #For speed fetch the epi part of the object
  epi <- obj$epi
  
  for (x in 1:obj$noX) {
    for (y in 1:obj$noY) {
      #Extract all events happening in the location (x,y)
      epi.loc <- epi[epi$x == x & epi$y == y,]
      #Have a look at the first event
      first   <- epi.loc[1,]
      first$t <- -1e-4
      first$S <- first$S + 1
      first$E <- first$E - 1
      #if (!(x==1 & y==1)) {
      #  first$E=0; first$S=first$S+1
      #} else {
      #  first$E=1; first$I=0
      #}
    
      epi.loc <- rbind(first,epi.loc)
      matplot(epi.loc$t,epi.loc[,c("E","S","I")],type="s",col=c(1,2,3),
              xlim=c(0,max(epi$t)),ylim=c(0,max(epi[,c("E","S","I")])+1),lty=1)
      title(paste("(",x,",",y,")"))
      if ((x==1) & (y==1)) {
        legend(max(epi$t)/2,max(epi[,c("E","S","I")]),
               c("E","S","I"),col=c(1,2,3),lty=1)
      }
    }
  }
  par(mfcol=c(1,1),mar=c(4,4,4,4))
}

######################################################################
# Plot the number of exposed, infectious and recovered summed
# over all cells.
#
# Params:
#  obj - the ladybug object containing the epi component
######################################################################
plot.epiglobal <- function(obj) {
  S <- numeric(dim(obj$epi)[1]+1)
  E <- numeric(dim(obj$epi)[1]+1)
  I <- numeric(dim(obj$epi)[1]+1)

  #Initial numbers
  S[1] <- sum(obj$setup$S+obj$setup$I)
  E[1] <- 0
  I[1] <- 0
  
  for (i in 1:dim(obj$epi)[1]) {
    type <- obj$epi[i,]$type
    #Action depends on type
    if (type == "E") {
      S[i+1] <- S[i] - 1 ; E[i+1] <- E[i] + 1 ; I[i+1] <- I[i]
    } else if (type == "I") {
      S[i+1] <- S[i]     ; E[i+1] <- E[i] - 1 ; I[i+1] <- I[i] + 1
    } else if (type == "R") {
      S[i+1] <- S[i]     ; E[i+1] <- E[i]     ; I[i+1] <- I[i] - 1
    } else if (type == "S") {
      S[i+1] <- S[i]     ; E[i+1] <- E[i]     ; I[i+1] <- I[i];
    }
  }
  #Plot
  matplot(c(0,obj$epi$t), cbind(S,E,I),type="s",xlab=expression(t),ylab="Indiviuals")
  legend(2/3*max(obj$epi$t),max(S,E,I),c("S","E","I"),lty=1,col=1:3)
}

######################################################################
# Helper function to draw a slice
######################################################################

draw.slice <- function(center,phi,r,col) {
  #Is there something to draw?
  if (abs(phi[2]-phi[1]) > 1e-2) {
    #Can we draw the entire thing in one slice?
    if (abs(phi[2]-phi[1] - 2*pi) < 0.01) {
      u <- seq(phi[1],phi[2],length=50)
      x <- r*1/6*cos(u) + center[1]
      y <- r*1/6*sin(u) + center[2]
      polygon(x,y,col=col)
    } else {
      #Include center
      u <- seq(phi[1],phi[2],length=50)
      x <- c(center[1],r*1/6*cos(u) + center[1],center[1])
      y <- c(center[2],r*1/6*sin(u) + center[2],center[2])
      polygon(x,y,col=col)
    }
  }
}

######################################################################
# This function illustrates the three multivariate time series
# (dimension: noX x noY) containing susceptibles, exposed, and
# infected using noOfPics pictures.
# For each picture a pdf file is created
#
# Params:
#  S - time series with the number of susceptibles in each cell
#  E - time series with the number of susceptibles in each cell
#  I - time series with the number of susceptibles in each cell
#  justInf - boolean telling whether to just show the infected
#  noOfPics - number of pictures for the animiation
#  PDF = boolean, whether to save the pictures as PDFs
######################################################################

show.animation <- function(obj,justInf=F,noOfPics=200,PDF=FALSE,addstr="") {
  #Extract relevant structures from the ladybug object
  S <- obj$ts$S
  E <- obj$ts$E
  I <- obj$ts$I
  name <- paste(obj$name,addstr,sep="")
  epi <- obj$epi
  
  #The time points where to show the picture (linear time)
  #If we would take epi$t we are not linear in time.
  tseq <- seq(0,max(epi$t),length=noOfPics)

  #largest possible number
  nmax <- max(S[,,])

  #Colors of <>, exposed, infected, susceptible, recovered
  colors <- c("","orange","red","green","gray")
  
  #Loop over all time points
  for (i in 1:length(tseq)) {
    #Find the idx in the multivariate time series corresponding
    #to t
    idx <- which.max(epi$t >= tseq[i]) + 1

    if (PDF) {
      fileName <- paste(name,"-",i,".pdf",sep="")
    
      pdf(fileName,height=5,width=5*(obj$noY/obj$noX))
      cat("Creating ",fileName,"\n")
    }
    #Set the par
    oldmar <- par()$mar
    #bottom left top right (
    par(mar=c(0,0,2,0))
    
    #Make the plot
    plot(NA,xlim=c(-1/6, (obj$noY-1)/3 + 1/6),ylim=c(-2/6,(obj$noX-1)/3+1/6),axes=F,ylab="x-coordinate",xlab="y-coordinate")
    for (x in 1:obj$noX) {
      for(y in 1:obj$noY) {
        #Show either only the infectious or all S,E,I,R
        if (justInf) {
          draw.slice(c((1/3)*(y-1),(obj$noX-x)*1/3),phi=c(0,2*pi),r=I[x,y,idx]/nmax,col="red")
        } else {
          #Draw all slices
          #Scale the circles according to the population
          r <- S[x,y,1]/nmax 
          #Compute the proportions of E,I,R,T
          angles <- c(E[x,y,idx],I[x,y,idx],S[x,y,idx],S[x,y,1]-E[x,y,idx]-I[x,y,idx]-S[x,y,idx])
          angles <- c(0,2*pi*cumsum(angles)/sum(angles))
          #
          for (j in 2:5) {
            draw.slice(c((1/3)*(y-1),(obj$noX-x)*1/3),phi=c(angles[j-1],angles[j]),r,col=colors[j])
          }
        }
      }
    }
    title(paste("t=",round(tseq[i]*10)/10))
    legend(-1/6,-1/6-1/20,c("Exposed","Infected","Susceptible","Recovered"),horiz=T,fill=c("orange","red","green","gray"),bty="n")
    #Turn the device off.
    if (PDF) {
      dev.off()
    }
  }
  #Done.
  par(mar=oldmar)
  invisible()
}


######################################################################
# Function to load the data from the <name>.all and <name>.data file
######################################################################
load.data <- function(name = "foobar") {
  #Read the data from the LadyBug <name>.all file
  epi <- read.table(paste(name,".all",sep=""),col.names=c("x","y","t","type"))
  #Determine the number of units in the setup
  noX <- max(epi$x)
  noY <- max(epi$y)
  noOfUnits <- noX * noY
  #Read the secondary information
  setup <- read.table(paste(name,".data",sep=""),skip=4,nrow=noOfUnits,col.names=c("u","x","y","S","I"))

  ##Create an S3 object containing the information
  obj <- list(name=name,epi=epi,noX=noX,noY=noY,noOfUnits=noOfUnits,setup=setup)
  class(obj) <- "ladybug"

  ##Init the epi part of the obj
  obj <- init(obj)
  
  
  ######################################################################
  #Generate 3 multivariate time series containing
  #the susceptibles, exposed, infectious
  ######################################################################
  obj <- init.spacetime(obj)

  #Return the result
  return(obj)
}

######################################################################
#Plot function for the ladybug object
#Params:
# pdf - create PDF files?
######################################################################
plot.ladybug <- function(obj,asPDF=TRUE,noOfPics=200,...) {
  ## Plot of all units
  if (asPDF) {
    pdf(paste(obj$name,"-gtraj.pdf",sep=""),...)
  }
  par(mar=c(4,4,2,2))
  plot.epiglobal(obj)
  if (asPDF) {
    dev.off()
  }
  
  #Local units
  if (asPDF) pdf(paste(obj$name,"-ltraj.pdf",sep=""),...)
  plot.epilocal(obj)
  if (asPDF) dev.off()
  
  #Show the inf time series
  show.animation(obj,noOfPics=noOfPics,PDF=asPDF)
  show.animation(obj,justInf=TRUE,noOfPics=noOfPics,PDF=asPDF,addstr="-2")
}


######################################################################
# Here is the action
######################################################################
##These are the data from the CSDA2005 talk
#sim <-  load.data("sim8x16")

#Just load the standard file name "foobar" made by the simSellke program
sim <-  load.data("foobar")

#Show it on the screen?
plot(sim,noOfPics=200,asPDF=F)
plot(sim,noOfPics=8,asPDF=T)

